@@ -111,7 +111,7 @@ def main(cmdlineargs=None, default_outfile="-"):
         rc, summary = atropos.multicore.run_parallel(
             reader, modifiers, filters, formatters, writers, options.threads,
             options.process_timeout, options.preserve_order, options.read_queue_size,
-            options.result_queue_size, not options.no_writer_process, options.worker_compression)
+            options.result_queue_size, not options.no_writer_process, options.compression)
     
     if rc != 0:
         sys.exit(rc)
@@ -393,27 +393,18 @@ def get_argument_parser():
     group.add_argument("--preserve-order", action="store_true", default=False,
         help="Preserve order of reads in input files (ignored if "
              "--no-writer-process is set). (no)")
-    group.add_argument("--process-timeout", type=int, default=30, metavar="SECONDS",
+    group.add_argument("--process-timeout", type=int, default=60, metavar="SECONDS",
         help="Number of seconds process should wait before escalating messages "
-             "to ERROR level. (30)")
-    group.add_argument("--parallel-environment", choices=("local", "cluster"), default="local",
-        help="Automatically set performance-tuning parameters based on the environment in which "
-             "the program is running. local: assumes fast file I/O and more limited memory; "
-             "cluster: assumes slower (i.e. network-based) file I/O and greater memory availablity. "
-             "(local)")
-    
-    group = parser.add_argument_group("Performance-tuning options")
+             "to ERROR level. (60)")
     group.add_argument("--batch-size", type=int, default=5000, metavar="SIZE",
         help="Number of records to process in each batch. (5000)")
     group.add_argument("--read-queue-size", type=int, default=None, metavar="SIZE",
-        help="Size of queue for batches of reads to be processed. (THREADS * 100 "
-             "in local mode, and THREADS * 1000 in cluster mode)")
+        help="Size of queue for batches of reads to be processed. (THREADS * 100)")
     group.add_argument("--result-queue-size", type=int, default=None, metavar="SIZE",
-        help="Size of queue for batches of results to be written. (THREADS * 100 "
-             "in local mode, and THREADS * 1000 in cluster mode)")
-    group.add_argument("--worker-compression", action="store_true", default=False,
-        help="Perform data compression in the worker processes rather than the "
-             "writer process. (no)")
+        help="Size of queue for batches of results to be written. (THREADS * 100)")
+    group.add_argument("--compression", choices=("worker", "writer"), default=None,
+        help="Where data compression should be performed. Defaults to 'writer' if "
+             "system-level compression can be used, otherwise defaults to 'worker'.")
     
     return parser
 
@@ -663,37 +654,38 @@ def validate_options(options, parser):
     options.max_reads = int_or_str(options.max_reads)
     
     if options.threads is not None:
+        if options.compression is None:
+            if options.no_writer_process:
+                options.compression = "worker"
+            else:
+                from atropos.compression import can_use_system_compression
+                options.compression = "writer" if can_use_system_compression() else "worker"
+        elif options.compression == "writer" and options.no_writer_process:
+            parser.error(" writer compression and --no-writer-process are mutually exclusive")
+        
         threads = options.threads
         if threads <= 0:
             threads = cpu_count()
-        # reserve one thread each for the reader and writer processes
-        threads = threads - (1 if options.no_writer_process else 2)
-        if threads <= 0:
-            logging.getLogger().warn("Using more than the available number of CPUs")
-            threads = 1
+        if threads < 2:
+            parser.error("At least two threads are required for multi-processing")
         options.threads = threads
         
         options.batch_size = int_or_str(options.batch_size)
         if options.process_timeout < 0:
             options.process_timeout = 0
         
-        # set default option values based on running environment
-        if options.parallel_environment == "local":
-            default_queue_size = options.threads * 100
-        elif options.parallel_environment == "cluster":
-            default_queue_size = options.threads * 1000
-            options.worker_compression = True
-        
-        # set queue sizes if necessary
+        # Set queue sizes if necessary.
+        # If we are using writer compression, the back-up will be in the result queue,
+        # otherwise it will be in the read queue.
         if options.read_queue_size is None:
-            options.read_queue_size = default_queue_size
+            options.read_queue_size = threads * (100 if options.compression == "writer" else 1000)
         elif options.read_queue_size > 0:
-            assert options.read_queue_size >= options.threads
+            assert options.read_queue_size >= threads
     
         if options.result_queue_size is None:
-            options.result_queue_size = default_queue_size
+            options.result_queue_size = threads * (100 if options.compression == "worker" else 1000)
         elif options.result_queue_size > 0:
-            assert options.result_queue_size > options.threads
+            assert options.result_queue_size >= threads
     
     return paired
 
